from abc import abstractmethod
from collections.abc import Sequence
from enum import Enum
from itertools import combinations
from typing import Optional

from pharia_inference_sdk.core import Input, Output
from pydantic import BaseModel

from pharia_studio_sdk.connectors.base.json_serializable import (
    SerializableDict,
)
from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import Example, ExpectedOutput
from pharia_studio_sdk.evaluation.evaluation.domain import (
    Evaluation,
    EvaluationOverview,
)
from pharia_studio_sdk.evaluation.evaluation.evaluation_repository import (
    EvaluationRepository,
)
from pharia_studio_sdk.evaluation.evaluation.evaluator.evaluator import (
    EvaluationLogic,
    Evaluator,
)
from pharia_studio_sdk.evaluation.run.domain import SuccessfulExampleOutput
from pharia_studio_sdk.evaluation.run.run_repository import RunRepository


class IncrementalEvaluationLogic(
    EvaluationLogic[Input, Output, ExpectedOutput, Evaluation]
):
    def __init__(self) -> None:
        super().__init__()
        self._previous_run_output_ids: list[set[str]] = []

    def set_previous_run_output_ids(
        self, previous_run_output_ids: list[set[str]]
    ) -> None:
        self._previous_run_output_ids = previous_run_output_ids

    def do_evaluate(
        self,
        example: Example[Input, ExpectedOutput],
        *output: SuccessfulExampleOutput[Output],
    ) -> Evaluation:
        """Executes the evaluation for this specific example.

        Responsible for comparing the input & expected output of a task to the
        actually generated output. The difference to the standard :class:`EvaluationLogic`'s `do_evaluate` is that
        this method will separate already processed evaluation from new ones before handing them over to
        `do_incremental_evaluate`.

        Args:
            example: Input data of :class:`Task` to produce the output.
            *output: Outputs of the :class:`Task`.

        Returns:
            :class:`Evaluation`: The metrics that come from the evaluated :class:`Task`.
        """
        already_evaluated_outputs = []
        for run_output_ids in self._previous_run_output_ids:
            already_evaluated_outputs.append(
                [
                    current_output
                    for current_output in output
                    if current_output.run_id in run_output_ids
                ]
            )

        return self.do_incremental_evaluate(
            example, list(output), already_evaluated_outputs
        )

    @abstractmethod
    def do_incremental_evaluate(
        self,
        example: Example[Input, ExpectedOutput],
        outputs: list[SuccessfulExampleOutput[Output]],
        already_evaluated_outputs: list[list[SuccessfulExampleOutput[Output]]],
    ) -> Evaluation:
        pass


class IncrementalEvaluator(Evaluator[Input, Output, ExpectedOutput, Evaluation]):
    """:class:`Evaluator` for evaluating additional runs on top of previous evaluations. Intended for use with :class:`IncrementalEvaluationLogic`.

    Args:
        dataset_repository: The repository with the examples that will be taken for the evaluation.
        run_repository: The repository of the runs to evaluate.
        evaluation_repository: The repository that will be used to store evaluation results.
        description: Human-readable description for the evaluator.
        incremental_evaluation_logic: The logic to use for evaluation.

    Generics:
        Input: Interface to be passed to the :class:`Task` that shall be evaluated.
        Output: Type of the output of the :class:`Task` to be evaluated.
        ExpectedOutput: Output that is expected from the run with the supplied input.
        Evaluation: Interface of the metrics that come from the evaluated :class:`Task`.
    """

    def __init__(
        self,
        dataset_repository: DatasetRepository,
        run_repository: RunRepository,
        evaluation_repository: EvaluationRepository,
        description: str,
        incremental_evaluation_logic: IncrementalEvaluationLogic[
            Input, Output, ExpectedOutput, Evaluation
        ],
    ) -> None:
        super().__init__(
            dataset_repository=dataset_repository,
            run_repository=run_repository,
            evaluation_repository=evaluation_repository,
            description=description,
            evaluation_logic=incremental_evaluation_logic,
        )
        self._evaluation_logic: IncrementalEvaluationLogic[
            Input, Output, ExpectedOutput, Evaluation
        ]

    def evaluate_additional_runs(
        self,
        *run_ids: str,
        previous_evaluation_ids: Optional[list[str]] = None,
        num_examples: Optional[int] = None,
        abort_on_error: bool = False,
        labels: Optional[set[str]] = None,
        metadata: Optional[SerializableDict] = None,
    ) -> EvaluationOverview:
        """Evaluate all runs while considering which runs have already been evaluated according to `previous_evaluation_id`.

        For each set of successful outputs in the referenced runs,
        :func:`EvaluationLogic.do_evaluate` is called and eval metrics are produced &
        stored in the provided :class:`EvaluationRepository`.

        Args:
            *run_ids: The runs to be evaluated. Each run is expected to have the same
                dataset as input (which implies their tasks have the same input-type)
                and their tasks have the same output-type. For each example in the
                dataset referenced by the runs the outputs of all runs are collected
                and if all of them were successful they are passed on to the implementation
                specific evaluation. The method compares all run of the provided ids to each other.
            previous_evaluation_ids: IDs of previous evaluation to consider
            num_examples: The number of examples which should be evaluated from the given runs.
                Always the first n runs stored in the evaluation repository. Defaults to None.
            abort_on_error: Flag to abort all evaluations when an error occurs. Defaults to False.
            labels: A list of labels for filtering. Defaults to an empty list.
            metadata: A dict for additional information about the evaluation overview. Defaults to an empty dict.

        Returns:
            EvaluationOverview: An overview of the evaluation. Individual :class:`Evaluation`s will not be
            returned but instead stored in the :class:`EvaluationRepository` provided in the
            __init__.
        """
        if metadata is None:
            metadata = dict()
        if labels is None:
            labels = set()
        previous_run_ids = []
        previous_evaluation_ids = previous_evaluation_ids or []

        for previous_evaluation_id in previous_evaluation_ids:
            prev_run_ids: set[str] = set()
            lineages = self.evaluation_lineages(previous_evaluation_id)
            for lineage in lineages:
                for output in lineage.outputs:
                    prev_run_ids.add(output.run_id)
            previous_run_ids.append(prev_run_ids)

        self._evaluation_logic.set_previous_run_output_ids(previous_run_ids)
        return super().evaluate_runs(
            *run_ids,
            num_examples=num_examples,
            abort_on_error=abort_on_error,
            labels=labels,
            metadata=metadata,
        )

    def evaluate_runs(
        self,
        *run_ids: str,
        num_examples: Optional[int] = None,
        abort_on_error: bool = False,
        skip_example_on_any_failure: bool = True,
        description: Optional[str] = None,
        labels: set[str] | None = None,
        metadata: SerializableDict | None = None,
    ) -> EvaluationOverview:
        if metadata is None:
            metadata = dict()
        if labels is None:
            labels = set()
        self._evaluation_logic.set_previous_run_output_ids([])
        return super().evaluate_runs(
            *run_ids,
            num_examples=num_examples,
            skip_example_on_any_failure=skip_example_on_any_failure,
            abort_on_error=abort_on_error,
            description=description,
        )


class MatchOutcome(str, Enum):
    A_WINS = "a_wins"
    DRAW = "draw"
    B_WINS = "b_wins"

    @property
    def payoff(self) -> tuple[float, float]:
        if self == self.A_WINS:
            return (1, 0)
        if self == self.DRAW:
            return (0.5, 0.5)
        return (0, 1)

    @staticmethod
    def from_rank_literal(rank: int) -> "MatchOutcome":
        match rank:
            case 1:
                return MatchOutcome.A_WINS
            case 2:
                return MatchOutcome.B_WINS
            case 3:
                return MatchOutcome.DRAW
            case _:
                raise ValueError(f"Got unexpected rank {rank}")


class ComparisonEvaluation(BaseModel):
    first_player: str
    second_player: str
    outcome: MatchOutcome


class Matches(BaseModel):
    comparison_evaluations: Sequence[ComparisonEvaluation]


class EloGradingInput(BaseModel):
    instruction: str
    first_completion: str
    second_completion: str


class EloEvaluationLogic(
    IncrementalEvaluationLogic[Input, Output, ExpectedOutput, Matches]
):
    def __init__(self) -> None:
        super().__init__()
        self._previous_run_output_ids: list[set[str]] = []

    def set_previous_run_output_ids(
        self, previous_run_output_ids: list[set[str]]
    ) -> None:
        self._previous_run_output_ids = previous_run_output_ids

    def do_incremental_evaluate(
        self,
        example: Example[Input, ExpectedOutput],
        outputs: list[SuccessfulExampleOutput[Output]],
        already_evaluated_outputs: list[list[SuccessfulExampleOutput[Output]]],
    ) -> Matches:
        pairs = combinations(outputs, 2)
        unique_pre_evaluated_runs: set[str] = set()

        for pre_run_output in already_evaluated_outputs:
            for current_output in pre_run_output:
                unique_pre_evaluated_runs.add(current_output.run_id)

        return Matches(
            comparison_evaluations=[
                ComparisonEvaluation(
                    first_player=player_a.run_id,
                    second_player=player_b.run_id,
                    outcome=self.grade(player_a, player_b, example),
                )
                for [player_a, player_b] in pairs
                if unique_pre_evaluated_runs is None
                or len(unique_pre_evaluated_runs) == 0
                or not (
                    player_a.run_id in unique_pre_evaluated_runs
                    and player_b.run_id in unique_pre_evaluated_runs
                )
            ]
        )

    @abstractmethod
    def grade(
        self,
        first: SuccessfulExampleOutput[Output],
        second: SuccessfulExampleOutput[Output],
        example: Example[Input, ExpectedOutput],
    ) -> MatchOutcome:
        """Returns a :class: `MatchOutcome` for the provided two contestants on the given example.

        Defines the use case specific logic how to determine the winner of the two provided outputs.


        Args:
            first: Instance of :class: `SuccessfulExampleOutut[Output]` of the first contestant in the comparison
            second: Instance of :class: `SuccessfulExampleOutut[Output]` of the second contestant in the comparison
            example: Datapoint of :class: `Example` on which the two outputs were generated

        Returns:
            Instance of :class: `MatchOutcome`
        """
        pass



import random
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from datetime import datetime
from itertools import combinations
from typing import Any, ClassVar, Optional
from uuid import uuid4

import argilla as rg
from pharia_inference_sdk.core import CompleteOutput, Input, InstructInput, Output
from pydantic import BaseModel

from pharia_studio_sdk.connectors.argilla.argilla_client import (
    ArgillaClient,
    ArgillaEvaluation,
    RecordData,
)
from pharia_studio_sdk.connectors.base.json_serializable import (
    SerializableDict,
)
from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import Example, ExpectedOutput
from pharia_studio_sdk.evaluation.evaluation.domain import (
    Evaluation,
    EvaluationOverview,
    ExampleEvaluation,
    FailedExampleEvaluation,
    PartialEvaluationOverview,
)
from pharia_studio_sdk.evaluation.evaluation.evaluator.async_evaluator import (
    AsyncEvaluationRepository,
    AsyncEvaluator,
)
from pharia_studio_sdk.evaluation.evaluation.evaluator.base_evaluator import (
    EvaluationLogicBase,
)
from pharia_studio_sdk.evaluation.evaluation.evaluator.incremental_evaluator import (
    ComparisonEvaluation,
    MatchOutcome,
)
from pharia_studio_sdk.evaluation.run.domain import SuccessfulExampleOutput
from pharia_studio_sdk.evaluation.run.run_repository import RunRepository


class RecordDataSequence(BaseModel):
    records: Sequence[RecordData]


class ArgillaEvaluationLogic(
    EvaluationLogicBase[Input, Output, ExpectedOutput, Evaluation], ABC
):
    def __init__(self, fields: Mapping[str, Any], questions: Sequence[Any]):
        self.fields = fields
        self.questions = questions

    @abstractmethod
    def to_record(
        self,
        example: Example[Input, ExpectedOutput],
        *output: SuccessfulExampleOutput[Output],
    ) -> RecordDataSequence:
        """This method is responsible for translating the `Example` and `Output` of the task to :class:`RecordData`.

        The specific format depends on the `fields`.


        Args:
            example: The example to be translated.
            *output: The output of the example that was run.

        Returns:
            A :class:`RecordDataSequence` that contains entries that should be evaluated in Argilla.
        """
        ...

    @abstractmethod
    def from_record(self, argilla_evaluation: ArgillaEvaluation) -> Evaluation:
        """This method takes the specific Argilla evaluation format and converts into a compatible :class:`Evaluation`.

        The format of argilla_evaluation.responses depends on the `questions` attribute.
        Each `name` of a question will be a key in the `argilla_evaluation.responses` mapping.

        Args:
            argilla_evaluation: Argilla-specific data for a single evaluation.

        Returns:
            An :class:`Evaluation` that contains all evaluation specific data.
        """


class ArgillaEvaluator(AsyncEvaluator[Input, Output, ExpectedOutput, Evaluation]):
    """Evaluator used to integrate with Argilla (https://github.com/argilla-io/argilla).

    Use this evaluator if you would like to easily do human eval.
    This evaluator runs a dataset and sends the input and output to Argilla to be evaluated.

    Arguments:
        dataset_repository: The repository with the examples that will be taken for the evaluation.
        run_repository: The repository of the runs to evaluate.
        evaluation_repository: The repository that will be used to store evaluation results.
        description: Human-readable description for the evaluator.
        evaluation_logic: The logic to use for evaluation.
        argilla_client: The client to interface with argilla.
        workspace_id: The argilla workspace id where datasets are created for evaluation.

    See the :class:`EvaluatorBase` for more information.
    """

    def __init__(
        self,
        dataset_repository: DatasetRepository,
        run_repository: RunRepository,
        evaluation_repository: AsyncEvaluationRepository,
        description: str,
        evaluation_logic: ArgillaEvaluationLogic[
            Input, Output, ExpectedOutput, Evaluation
        ],
        argilla_client: ArgillaClient,
        workspace_id: str,
    ) -> None:
        super().__init__(
            dataset_repository,
            run_repository,
            evaluation_repository,
            description,
            evaluation_logic,
        )
        self._client = argilla_client
        self._workspace_id = workspace_id
        self._evaluation_logic: ArgillaEvaluationLogic[
            Input, Output, ExpectedOutput, Evaluation
        ]
        self._evaluation_repository: AsyncEvaluationRepository

    def submit(
        self,
        *run_ids: str,
        num_examples: Optional[int] = None,
        dataset_name: Optional[str] = None,
        abort_on_error: bool = False,
        skip_example_on_any_failure: bool = True,
        labels: Optional[set[str]] = None,
        metadata: Optional[SerializableDict] = None,
    ) -> PartialEvaluationOverview:
        if metadata is None:
            metadata = dict()
        if labels is None:
            labels = set()
        argilla_dataset_id = self._client.create_dataset(
            self._workspace_id,
            dataset_name if dataset_name else str(uuid4()),
            fields=list(self._evaluation_logic.fields.values()),
            questions=self._evaluation_logic.questions,
        )

        run_overviews = self._load_run_overviews(*run_ids)
        submit_count = 0
        for example, outputs in self._retrieve_eval_logic_input(
            run_overviews,
            skip_example_on_any_failure=skip_example_on_any_failure,
            num_examples=num_examples,
        ):
            record_sequence = self._evaluation_logic.to_record(example, *outputs)
            for record in record_sequence.records:
                try:
                    self._client.add_record(argilla_dataset_id, record)
                    submit_count += 1
                except Exception as e:
                    if abort_on_error:
                        raise e
                    evaluation = FailedExampleEvaluation.from_exception(e)
                    self._evaluation_repository.store_example_evaluation(
                        ExampleEvaluation(
                            evaluation_id=argilla_dataset_id,
                            example_id=example.id,
                            result=evaluation,
                        )
                    )
                    print(
                        f"Uploading a record to argilla failed with the following error:\n{e}"
                    )

        partial_overview = PartialEvaluationOverview(
            run_overviews=frozenset(run_overviews),
            id=argilla_dataset_id,
            start_date=datetime.now(),
            submitted_evaluation_count=submit_count,
            description=self.description,
            labels=labels,
            metadata=metadata,
        )

        self._evaluation_repository.store_partial_evaluation_overview(partial_overview)
        return partial_overview

    def retrieve(
        self,
        partial_evaluation_id: str,
    ) -> EvaluationOverview:
        partial_overview = self._evaluation_repository.partial_evaluation_overview(
            partial_evaluation_id
        )
        if not partial_overview:
            raise ValueError(
                f"Partial overview for evaluation id {partial_evaluation_id} not found."
            )

        example_evaluations = [
            ExampleEvaluation(
                evaluation_id=partial_evaluation_id,
                example_id=example_evaluation.example_id,
                # cast to Evaluation because mypy thinks ArgillaEvaluation cannot be Evaluation
                result=self._evaluation_logic.from_record(example_evaluation),
            )
            for example_evaluation in self._client.evaluations(partial_evaluation_id)
        ]
        evaluations = sorted(example_evaluations, key=lambda i: i.example_id)

        for evaluation in evaluations:
            self._evaluation_repository.store_example_evaluation(evaluation)
        num_failed_evaluations = len(
            self._evaluation_repository.failed_example_evaluations(
                partial_evaluation_id, self.evaluation_type()
            )
        )
        num_not_yet_evaluated_evals = partial_overview.submitted_evaluation_count - len(
            evaluations
        )

        overview = EvaluationOverview(
            run_overviews=partial_overview.run_overviews,
            id=partial_evaluation_id,
            start_date=partial_overview.start_date,
            description=partial_overview.description,
            end_date=datetime.now(),
            successful_evaluation_count=len(evaluations),
            failed_evaluation_count=num_not_yet_evaluated_evals
            + num_failed_evaluations,
            labels=partial_overview.labels,
            metadata=partial_overview.metadata,
        )
        self._evaluation_repository.store_evaluation_overview(overview)
        return overview


class InstructComparisonArgillaEvaluationLogic(
    ArgillaEvaluationLogic[InstructInput, CompleteOutput, None, ComparisonEvaluation]
):
    KEY_INSTRUCTION = "instruction"
    KEY_INPUT = "input"
    KEY_RESPONSE_1 = "first"
    KEY_RESPONSE_2 = "second"
    KEY_QUESTION = "winner"
    OPTIONS: ClassVar[list[int]] = [1, 2, 3]

    def __init__(
        self,
        high_priority_runs: Optional[frozenset[str]] = None,
    ) -> None:
        self._high_priority_runs = high_priority_runs
        super().__init__(
            fields={
                "KEY_INSTRUCTION": rg.TextField(
                    name=self.KEY_INSTRUCTION, title="Instruction"
                ),
                "KEY_INPUT": rg.TextField(name=self.KEY_INPUT, title="Input"),
                "KEY_RESPONSE_1": rg.TextField(
                    name=self.KEY_RESPONSE_1, title="Response 1"
                ),
                "KEY_RESPONSE_2": rg.TextField(
                    name=self.KEY_RESPONSE_2, title="Response 2"
                ),
            },
            questions=[
                rg.RatingQuestion(
                    name=self.KEY_QUESTION,
                    title="Which response is better?",
                    description="1: The first completion is better.\n2: The second completion is better.\n3: They are both equally good.",
                    values=self.OPTIONS,
                )
            ],
        )

    def to_record(
        self,
        example: Example[InstructInput, None],
        *outputs: SuccessfulExampleOutput[CompleteOutput],
    ) -> RecordDataSequence:
        pairs = combinations(outputs, 2)
        return RecordDataSequence(
            records=[
                self._create_record_data(example, first, second)
                for [first, second] in pairs
                if self._high_priority_runs is None
                or any(
                    run_id in self._high_priority_runs
                    for run_id in [first.run_id, second.run_id]
                )
            ]
        )

    def _create_record_data(
        self,
        example: Example[InstructInput, None],
        first: SuccessfulExampleOutput[CompleteOutput],
        second: SuccessfulExampleOutput[CompleteOutput],
    ) -> RecordData:
        if random.choice([True, False]):
            first, second = second, first
        return RecordData(
            content={
                self.fields["KEY_INSTRUCTION"].name: example.input.instruction,
                self.fields["KEY_INPUT"].name: example.input.input or "",
                self.fields["KEY_RESPONSE_1"].name: first.output.completion,
                self.fields["KEY_RESPONSE_2"].name: second.output.completion,
            },
            example_id=example.id,
            metadata={
                self.fields["KEY_RESPONSE_1"].name: first.run_id,
                self.fields["KEY_RESPONSE_2"].name: second.run_id,
            },
        )

    def from_record(
        self, argilla_evaluation: ArgillaEvaluation
    ) -> ComparisonEvaluation:
        return ComparisonEvaluation(
            first_player=argilla_evaluation.metadata[
                self.fields["KEY_RESPONSE_1"].name
            ],
            second_player=argilla_evaluation.metadata[
                self.fields["KEY_RESPONSE_2"].name
            ],
            outcome=MatchOutcome.from_rank_literal(
                int(argilla_evaluation.responses["winner"])
            ),
        )



import traceback
from datetime import datetime
from typing import Generic

from pharia_inference_sdk.core.task import Output
from pydantic import BaseModel
from rich.tree import Tree

from pharia_studio_sdk.connectors.base.json_serializable import SerializableDict


class FailedExampleRun(BaseModel):
    """Captures an exception raised when running a single example with a :class:`Task`.

    Attributes:
        error_message: String-representation of the exception.
    """

    error_message: str

    @staticmethod
    def from_exception(exception: Exception) -> "FailedExampleRun":
        return FailedExampleRun(
            error_message=f"{type(exception).__qualname__}: {exception}\n{traceback.format_exc()}"
        )


class ExampleOutput(BaseModel, Generic[Output]):
    """Output of a single evaluated :class:`Example`.

    Created to persist the output (including failures) of an individual example in the repository.

    Attributes:
        run_id: Identifier of the run that created the output.
        example_id: Identifier of the :class:`Example` that provided the input for generating the output.
        output: Generated when running the :class:`Task`. When the running the task
            failed this is an :class:`FailedExampleRun`.

    Generics:
        Output: Interface of the output returned by the task.
    """

    run_id: str
    example_id: str
    output: Output | FailedExampleRun

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return (
            f"Example ID={self.example_id}\n"
            f"Related Run ID={self.run_id}\n"
            f'Output="{self.output}"\n'
        )

    def _rich_render(self, skip_example_id: bool = False) -> Tree:
        tree = Tree(f"Output: {self.run_id}")
        if not skip_example_id:
            tree.add(f"Example ID: {self.example_id}")
        tree.add(str(self.output))
        return tree


class SuccessfulExampleOutput(BaseModel, Generic[Output]):
    """Successful output of a single evaluated :class:`Example`.

    Attributes:
        run_id: Identifier of the run that created the output.
        example_id: Identifier of the :class:`Example`.
        output: Generated when running the :class:`Task`. This represent only
            the output of an successful run.

    Generics:
        Output: Interface of the output returned by the task.
    """

    run_id: str
    example_id: str
    output: Output

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return (
            f"Run ID = {self.run_id}\n"
            f"Example ID = {self.example_id}\n"
            f'Output = "{self.output}"\n'
        )


class RunOverview(BaseModel, frozen=True):
    """Overview of the run of a :class:`Task` on a dataset.

    Attributes:
        dataset_id: Identifier of the dataset run.
        id: The unique identifier of this run.
        start: The time when the run was started
        end: The time when the run ended
        failed_example_count: The number of examples where an exception was raised when running the task.
        successful_example_count: The number of examples that where successfully run.
        description: Human-readable of the runner that run the task.
        labels: Labels for filtering runs. Defaults to empty list.
        metadata: Additional information about the run. Defaults to empty dict.
    """

    dataset_id: str
    id: str
    start: datetime
    end: datetime
    failed_example_count: int
    successful_example_count: int
    description: str
    labels: set[str] = set()
    metadata: SerializableDict = dict()

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return (
            f"Run Overview ID = {self.id}\n"
            f"Dataset ID = {self.dataset_id}\n"
            f"Start time = {self.start}\n"
            f"End time = {self.end}\n"
            f"Failed example count = {self.failed_example_count}\n"
            f"Successful example count = {self.successful_example_count}\n"
            f'Description = "{self.description}"\n'
            f'Labels = "{self.labels}"\n'
            f'Metadata = "{self.metadata}"\n'
        )

    def __hash__(self) -> int:
        return hash(self.id)



from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed
from inspect import get_annotations
from itertools import islice
from typing import Any, Generic, Optional, cast
from uuid import uuid4

from dict_hash import dict_hash
from pharia_inference_sdk.core import (
    CompositeTracer,
    Input,
    NoOpTracer,
    Output,
    Task,
    Tracer,
    utc_now,
)
from tqdm import tqdm

from pharia_studio_sdk.connectors.base.json_serializable import (
    SerializableDict,
)
from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import Example, ExpectedOutput
from pharia_studio_sdk.evaluation.infrastructure.repository_navigator import (
    RepositoryNavigator,
    RunLineage,
)
from pharia_studio_sdk.evaluation.run.domain import (
    ExampleOutput,
    FailedExampleRun,
    RunOverview,
)
from pharia_studio_sdk.evaluation.run.run_repository import RunRepository


class Runner(Generic[Input, Output]):
    def __init__(
        self,
        task: Task[Input, Output],
        dataset_repository: DatasetRepository,
        run_repository: RunRepository,
        description: str,
    ) -> None:
        self._task = task
        self._run_repository = run_repository
        self._dataset_repository = dataset_repository
        self.description = description

    def output_type(self) -> type[Output]:
        """Returns the type of the evaluated task's output.

        This can be used to retrieve properly typed outputs of an evaluation run
        from a :class:`RunRepository`

        Returns:
            the type of the evaluated task's output.
        """
        try:
            output_type = get_annotations(self._task.do_run)["return"]
        except KeyError:
            raise TypeError(
                f"Task of type {type(self._task)} must have a type-hint for the return value of do_run to detect the output_type. "
                f"Alternatively overwrite output_type() in {type(self)}"
            ) from None
        return cast(type[Output], output_type)

    def input_type(self) -> type[Input]:
        try:
            input_type = get_annotations(self._task.do_run)["input"]
        except KeyError:
            raise TypeError(
                f"Task of type {type(self._task)} must have a type-hint for the input value of do_run to detect the input_type. "
                f"Alternatively overwrite input_type() in {type(self)}"
            ) from None
        return cast(type[Input], input_type)

    def _run_hash(self, dataset_id: str, run_description: str) -> str:
        return str(hash(dataset_id + self.description + run_description))

    def run_dataset(
        self,
        dataset_id: str,
        tracer: Optional[Tracer] = None,
        num_examples: Optional[int] = None,
        abort_on_error: bool = False,
        max_workers: int = 10,
        description: Optional[str] = None,
        trace_examples_individually: bool = True,
        labels: Optional[set[str]] = None,
        metadata: Optional[SerializableDict] = None,
        resume_from_recovery_data: bool = False,
    ) -> RunOverview:
        """Generates all outputs for the provided dataset.

        Will run each :class:`Example` provided in the dataset through the :class:`Task`.

        Args:
            dataset_id: The id of the dataset to generate output for. Consists of examples, each
                with an :class:`Input` and an :class:`ExpectedOutput` (can be None).
            tracer: An optional :class:`Tracer` to trace all the runs from each example.
                Use `trace_examples_individually` to trace each example with a dedicated tracer individually.
            num_examples: An optional int to specify how many examples from the dataset should be run.
                Always the first n examples will be taken.
            abort_on_error: Flag to abort all run when an error occurs. Defaults to False.
            max_workers: Number of examples that can be evaluated concurrently. Defaults to 10.
            description: An optional description of the run. Defaults to None.
            trace_examples_individually: Flag to create individual tracers for each example. Defaults to True.
            labels: A list of labels for filtering. Defaults to an empty list.
            metadata: A dict for additional information about the run overview. Defaults to an empty dict.
            resume_from_recovery_data: Flag to resume if execution failed previously.

        Returns:
            An overview of the run. Outputs will not be returned but instead stored in the
            :class:`RunRepository` provided in the __init__.
        """
        if labels is None:
            labels = set()
        if metadata is None:
            metadata = dict()

        run_id = str(uuid4())
        tmp_hash = self._run_hash(dataset_id, description or "")

        recovery_data = self._run_repository.finished_examples(tmp_hash)
        finished_examples: frozenset[str] = frozenset()
        if recovery_data is not None and resume_from_recovery_data:
            run_id = recovery_data.run_id
            finished_examples = frozenset(recovery_data.finished_examples)
        else:
            self._run_repository.create_temporary_run_data(tmp_hash, run_id)

        examples = self._dataset_repository.examples(
            dataset_id,
            self.input_type(),
            Any,  # type: ignore
            examples_to_skip=finished_examples,
        )
        if examples is None:
            raise ValueError(f"Dataset with id {dataset_id} not found")
        if num_examples:
            examples = islice(examples, num_examples)

        def run(
            example: Example[Input, ExpectedOutput],
        ) -> None:
            if trace_examples_individually:
                example_tracer = self._run_repository.create_tracer_for_example(
                    run_id, example.id
                )
                if tracer:
                    example_tracer = CompositeTracer([example_tracer, tracer])
            elif tracer:
                example_tracer = tracer
            else:
                example_tracer = NoOpTracer()

            output: Output | FailedExampleRun
            try:
                output = self._task.run(example.input, example_tracer)
            except Exception as e:
                if abort_on_error:
                    raise e
                print(
                    f'FAILED RUN: example "{example.id}", {type(e).__qualname__}: "{e}"'
                )
                output = FailedExampleRun.from_exception(e)

            self._run_repository.store_example_output_parallel(
                tmp_hash,
                ExampleOutput[Output](
                    run_id=run_id, example_id=example.id, output=output
                ),
            )
            self._run_repository.temp_store_finished_example(tmp_hash, example.id)

        start = utc_now()

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(run, example) for example in examples]  # type: ignore
            with tqdm(total=len(futures)) as pbar:
                pbar.set_description("Running Task")
                for future in as_completed(futures):
                    future.result()
                    pbar.update(1)
        self._run_repository.delete_temporary_run_data(tmp_hash)

        full_description = (
            self.description + " : " + description if description else self.description
        )

        successful = 0
        failed = 0
        for example_output in self._run_repository.example_outputs(
            run_id, self.output_type()
        ):
            if isinstance(example_output.output, FailedExampleRun):
                failed += 1
            else:
                successful += 1

        run_overview = RunOverview(
            dataset_id=dataset_id,
            id=run_id,
            start=start,
            end=utc_now(),
            failed_example_count=failed,
            successful_example_count=successful,
            description=full_description,
            labels=labels,
            metadata=metadata,
        )

        self._run_repository.store_run_overview(run_overview)
        return run_overview

    def failed_runs(
        self, run_id: str, expected_output_type: type[ExpectedOutput]
    ) -> Iterable[RunLineage[Input, ExpectedOutput, Output]]:
        """Returns the `RunLineage` objects for all failed example runs that belong to the given run ID.

        Args:
            run_id: The ID of the run overview
            expected_output_type: Type of output that the `Task` returned in :func:`Task.do_run`

        Returns:
            :class:`Iterable` of :class:`RunLineage`s.
        """
        failed_example_outputs = self._run_repository.failed_example_outputs(
            run_id, output_type=self.output_type()
        )
        lineages = (
            self.run_lineage(run_id, output.example_id, expected_output_type)
            for output in failed_example_outputs
        )
        return (lineage for lineage in lineages if lineage is not None)

    def run_lineages(
        self,
        run_id: str,
        expected_output_type: type[ExpectedOutput],
    ) -> Iterable[RunLineage[Input, ExpectedOutput, Output]]:
        """Wrapper for `RepositoryNavigator.run_lineages`.

        Args:
            run_id: The id of the run
            expected_output_type: The type of the expected output as defined by the :class:`Example`

        Returns:
            An iterator over all :class:`RunLineage`s for the given run id.
        """
        navigator = RepositoryNavigator(self._dataset_repository, self._run_repository)
        return navigator.run_lineages(
            run_id=run_id,
            input_type=self.input_type(),
            expected_output_type=expected_output_type,
            output_type=self.output_type(),
        )

    def run_lineage(
        self,
        run_id: str,
        example_id: str,
        expected_output_type: type[ExpectedOutput],
    ) -> RunLineage[Input, ExpectedOutput, Output] | None:
        """Wrapper for `RepositoryNavigator.run_lineage`.

        Args:
            run_id: The id of the run
            example_id: The id of the example of interest
            expected_output_type: The type of the expected output as defined by the :class:`Example`

        Returns:
            The :class:`RunLineage` for the given run id and example id, `None` if the example or an output for the example does not exist.
        """
        navigator = RepositoryNavigator(self._dataset_repository, self._run_repository)
        return navigator.run_lineage(
            run_id=run_id,
            example_id=example_id,
            input_type=self.input_type(),
            expected_output_type=expected_output_type,
            output_type=self.output_type(),
        )

    def run_is_already_computed(
        self,
        metadata: SerializableDict,
    ) -> bool:
        """Checks if a run with the given metadata has already been computed.

        Args:
            metadata: The metadata dictionary to check.

        Returns:
            True if a run with the same metadata has already been computed. False otherwise.
        """
        previous_runs = {
            dict_hash(run_overview.metadata)
            for run_overview in self._run_repository.run_overviews()
        }
        return dict_hash(metadata) in previous_runs



from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, final

from pharia_inference_sdk.core import Input, Output, utc_now
from tqdm import tqdm

from pharia_studio_sdk.connectors.base.json_serializable import (
    SerializableDict,
)
from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import Example, ExpectedOutput
from pharia_studio_sdk.evaluation.evaluation.domain import (
    Evaluation,
    EvaluationOverview,
    ExampleEvaluation,
    FailedExampleEvaluation,
)
from pharia_studio_sdk.evaluation.evaluation.evaluation_repository import (
    EvaluationRepository,
)
from pharia_studio_sdk.evaluation.evaluation.evaluator.base_evaluator import (
    EvaluationLogicBase,
    EvaluatorBase,
)
from pharia_studio_sdk.evaluation.run.domain import SuccessfulExampleOutput
from pharia_studio_sdk.evaluation.run.run_repository import RunRepository


class EvaluationLogic(
    ABC, EvaluationLogicBase[Input, Output, ExpectedOutput, Evaluation]
):
    @abstractmethod
    def do_evaluate(
        self,
        example: Example[Input, ExpectedOutput],
        *output: SuccessfulExampleOutput[Output],
    ) -> Evaluation:
        """Executes the evaluation for this specific example.

        Responsible for comparing the input & expected output of a task to the
        actually generated output.

        Args:
            example: Input data of :class:`Task` to produce the output.
            *output: Output of the :class:`Task`.

        Returns:
            The metrics that come from the evaluated :class:`Task`.
        """
        pass


class SingleOutputEvaluationLogic(
    EvaluationLogic[Input, Output, ExpectedOutput, Evaluation]
):
    @final
    def do_evaluate(
        self,
        example: Example[Input, ExpectedOutput],
        *output: SuccessfulExampleOutput[Output],
    ) -> Evaluation:
        assert len(output) == 1
        return self.do_evaluate_single_output(example, output[0].output)

    @abstractmethod
    def do_evaluate_single_output(
        self, example: Example[Input, ExpectedOutput], output: Output
    ) -> Evaluation:
        pass


class Evaluator(EvaluatorBase[Input, Output, ExpectedOutput, Evaluation]):
    """Evaluator designed for most evaluation tasks. Only supports synchronous evaluation.

    See the :class:`EvaluatorBase` for more information.
    """

    def __init__(
        self,
        dataset_repository: DatasetRepository,
        run_repository: RunRepository,
        evaluation_repository: EvaluationRepository,
        description: str,
        evaluation_logic: EvaluationLogic[Input, Output, ExpectedOutput, Evaluation],
    ) -> None:
        super().__init__(
            dataset_repository,
            run_repository,
            evaluation_repository,
            description,
            evaluation_logic,
        )
        self._evaluation_logic: EvaluationLogic[
            Input, Output, ExpectedOutput, Evaluation
        ]

    def evaluate_runs(
        self,
        *run_ids: str,
        num_examples: Optional[int] = None,
        abort_on_error: bool = False,
        skip_example_on_any_failure: bool = True,
        description: Optional[str] = None,
        labels: Optional[set[str]] = None,
        metadata: Optional[SerializableDict] = None,
    ) -> EvaluationOverview:
        """Evaluates all generated outputs in the run.

        For each set of successful outputs in the referenced runs,
        :func:`EvaluationLogic.do_evaluate` is called and eval metrics are produced &
        stored in the provided :class:`EvaluationRepository`.

        Args:
            *run_ids: The runs to be evaluated. Each run is expected to have the same
                dataset as input (which implies their tasks have the same input-type)
                and their tasks have the same output-type. For each example in the
                dataset referenced by the runs the outputs of all runs are collected
                and if all of them were successful they are passed on to the implementation
                specific evaluation. The method compares all run of the provided ids to each other.
            num_examples: The number of examples which should be evaluated from the given runs.
                Always the first n runs stored in the evaluation repository. Defaults to None.
            abort_on_error: Flag to abort all evaluations when an error occurs. Defaults to False.
            skip_example_on_any_failure: Flag to skip evaluation on any example for which at least one run fails. Defaults to True.
            description: Optional description of the evaluation. Defaults to None.
            labels: A list of labels for filtering. Defaults to an empty list.
            metadata: A dict for additional information about the evaluation overview. Defaults to an empty dict.

        Returns:
            EvaluationOverview: An overview of the evaluation. Individual :class:`Evaluation`s will not be
            returned but instead stored in the :class:`EvaluationRepository` provided in the
            __init__.
        """
        if metadata is None:
            metadata = dict()
        if labels is None:
            labels = set()
        start = utc_now()
        run_overviews = self._load_run_overviews(*run_ids)
        eval_id = self._evaluation_repository.initialize_evaluation()

        with ThreadPoolExecutor(max_workers=10) as executor:
            example_evaluations = list(  # the list is needed to consume the iterator returned from the executor.map
                tqdm(
                    executor.map(
                        lambda args: self.evaluate(
                            args[0], eval_id, abort_on_error, *args[1]
                        ),
                        self._retrieve_eval_logic_input(
                            run_overviews,
                            skip_example_on_any_failure=skip_example_on_any_failure,
                            num_examples=num_examples,
                        ),
                    ),
                    desc="Evaluating",
                )
            )

        failed_evaluation_count = sum(
            isinstance(example_evaluation, FailedExampleEvaluation)
            for example_evaluation in example_evaluations
        )

        successful_evaluation_count = len(example_evaluations) - failed_evaluation_count
        full_description = (
            self.description + " : " + description if description else self.description
        )
        overview = EvaluationOverview(
            run_overviews=frozenset(run_overviews),
            id=eval_id,
            start_date=start,
            end_date=utc_now(),
            successful_evaluation_count=successful_evaluation_count,
            failed_evaluation_count=failed_evaluation_count,
            description=full_description,
            labels=labels,
            metadata=metadata,
        )
        self._evaluation_repository.store_evaluation_overview(overview)

        return overview

    @final
    def evaluate(
        self,
        example: Example[Input, ExpectedOutput],
        evaluation_id: str,
        abort_on_error: bool,
        *example_outputs: SuccessfulExampleOutput[Output],
    ) -> Evaluation | FailedExampleEvaluation:
        try:
            result: Evaluation | FailedExampleEvaluation = (
                self._evaluation_logic.do_evaluate(
                    example,
                    *example_outputs,
                )
            )
        except Exception as e:
            if abort_on_error:
                raise e
            print(
                f'FAILED EVALUATION: example "{example.id}", {type(e).__qualname__}: "{e}"'
            )
            result = FailedExampleEvaluation.from_exception(e)
        self._evaluation_repository.store_example_evaluation(
            ExampleEvaluation(
                evaluation_id=evaluation_id, example_id=example.id, result=result
            )
        )

        return result



from pathlib import Path

import huggingface_hub

from pharia_studio_sdk.evaluation.infrastructure.file_system_based_repository import (
    FileSystemBasedRepository,
)


class HuggingFaceRepository(FileSystemBasedRepository):
    """HuggingFace base repository."""

    _REPO_TYPE = "dataset"
    _ROOT_DIRECTORY_PREFIX_ = "datasets"  # HuggingFace API root directory

    @staticmethod
    def path_to_str(path: Path) -> str:
        return path.as_posix()

    def __init__(self, repository_id: str, token: str, private: bool) -> None:
        """Create a HuggingFace repository.

        Creates a corresponding repository and initializes the file system.

        Args:
            repository_id: The HuggingFace namespace and repository name, separated by a "/".
            token: The HuggingFace authentication token.
            private: Whether the dataset repository should be private.
        """
        assert repository_id[-1] != "/"
        self.create_repository(repository_id, token, private)

        file_system = huggingface_hub.HfFileSystem(token=token)
        root_directory = Path(f"{self._ROOT_DIRECTORY_PREFIX_}/{repository_id}")

        super().__init__(file_system, root_directory)
        self._repository_id = repository_id
        # the file system is assigned in super init but this fixes the typing
        self._file_system: huggingface_hub.HfFileSystem

    def create_repository(self, repository_id: str, token: str, private: bool) -> None:
        huggingface_hub.create_repo(
            repo_id=repository_id,
            token=token,
            repo_type=self._REPO_TYPE,
            private=private,
            exist_ok=True,
        )

    def delete_repository(self) -> None:
        huggingface_hub.delete_repo(
            repo_id=self._repository_id,
            token=self._file_system.token,
            repo_type=self._REPO_TYPE,
            missing_ok=True,
        )



from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from typing import Optional
from uuid import uuid4

from pharia_inference_sdk.core import JsonSerializer
from pydantic import BaseModel

from pharia_studio_sdk.evaluation.evaluation.domain import (
    Evaluation,
    EvaluationOverview,
    ExampleEvaluation,
    FailedExampleEvaluation,
)


class SerializedExampleEvaluation(BaseModel):
    """A json-serialized evaluation of a single example in a dataset.

    Attributes:
        evaluation_id: ID of the linked :class:`EvaluationOverview`.
        example_id: ID of the :class:`ExampleEvaluation` this evaluation was created for.
        is_exception: Will be `True` if an exception occurred during the evaluation.
        json_result: The actual serialized evaluation result.
    """

    evaluation_id: str
    example_id: str
    is_exception: bool
    json_result: str

    @classmethod
    def from_example_result(
        cls, result: ExampleEvaluation[Evaluation]
    ) -> "SerializedExampleEvaluation":
        return cls(
            evaluation_id=result.evaluation_id,
            json_result=JsonSerializer(root=result.result).model_dump_json(),
            is_exception=isinstance(result.result, FailedExampleEvaluation),
            example_id=result.example_id,
        )

    def to_example_result(
        self, evaluation_type: type[Evaluation]
    ) -> ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]:
        expected_result_type = (
            FailedExampleEvaluation if self.is_exception else evaluation_type
        )
        return ExampleEvaluation(
            evaluation_id=self.evaluation_id,
            example_id=self.example_id,
            result=expected_result_type.model_validate_json(self.json_result),
        )


class EvaluationRepository(ABC):
    """Base evaluation repository interface.

    Provides methods to store and load evaluation results:
        :class:`EvaluationOverview`s and :class:`ExampleEvaluation`.
    An :class:`EvaluationOverview` is created from and is linked (by its ID)
        to multiple :class:`ExampleEvaluation`s.
    """

    def initialize_evaluation(self) -> str:
        """Initializes an :class:`EvaluationOverview` and returns its ID.

        If no extra logic is required for the initialization, this function just returns a UUID as string.
        In other cases (e.g., when a dataset has to be created in an external repository),
        this method is responsible for implementing this logic and returning the created ID.

        Returns:
            The created ID.
        """
        return str(uuid4())

    @abstractmethod
    def store_evaluation_overview(
        self, evaluation_overview: EvaluationOverview
    ) -> None:
        """Stores an :class:`EvaluationOverview`.

        Args:
            evaluation_overview: The overview to be persisted.
        """
        ...

    @abstractmethod
    def evaluation_overview(self, evaluation_id: str) -> Optional[EvaluationOverview]:
        """Returns an :class:`EvaluationOverview` for the given ID.

        Args:
            evaluation_id: ID of the evaluation overview to retrieve.

        Returns:
            :class:`EvaluationOverview` if it was found, `None` otherwise.
        """
        ...

    def evaluation_overviews(self) -> Iterable[EvaluationOverview]:
        """Returns all :class:`EvaluationOverview`s sorted by their ID.

        Yields:
            :class:`EvaluationOverview`s.
        """
        for eval_id in self.evaluation_overview_ids():
            evaluation_overview = self.evaluation_overview(eval_id)
            if evaluation_overview is not None:
                yield evaluation_overview

    @abstractmethod
    def evaluation_overview_ids(self) -> Sequence[str]:
        """Returns sorted IDs of all stored :class:`EvaluationOverview`s.

        Returns:
            A :class:`Sequence` of the :class:`EvaluationOverview` IDs.
        """
        ...

    @abstractmethod
    def store_example_evaluation(
        self, example_evaluation: ExampleEvaluation[Evaluation]
    ) -> None:
        """Stores an :class:`ExampleEvaluation`.

        Args:
            example_evaluation: The example evaluation to be persisted.
        """

    @abstractmethod
    def example_evaluation(
        self, evaluation_id: str, example_id: str, evaluation_type: type[Evaluation]
    ) -> Optional[
        ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]
    ]:
        """Returns an :class:`ExampleEvaluation` for the given evaluation overview ID and example ID.

        Args:
            evaluation_id: ID of the linked evaluation overview.
            example_id: ID of the example evaluation to retrieve.
            evaluation_type: Type of example evaluations that the `Evaluator` returned
                in :func:`Evaluator.do_evaluate`

        Returns:
            :class:`ExampleEvaluation` if it was found, `None` otherwise.
        """
        ...

    @abstractmethod
    def example_evaluations(
        self, evaluation_id: str, evaluation_type: type[Evaluation]
    ) -> Sequence[
        ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]
    ]:
        """Returns all :class:`ExampleEvaluation`s for the given evaluation overview ID sorted by their example ID.

        Args:
            evaluation_id: ID of the corresponding evaluation overview.
            evaluation_type: Type of evaluations that the :class:`Evaluator` returned
                in :func:`Evaluator.do_evaluate`.

        Returns:
            A :class:`Sequence` of :class:`ExampleEvaluation`s.
        """
        ...

    def successful_example_evaluations(
        self, evaluation_id: str, evaluation_type: type[Evaluation]
    ) -> Sequence[ExampleEvaluation[Evaluation]]:
        """Returns all successful :class:`ExampleEvaluation`s for the given evaluation overview ID sorted by their example ID.

        Args:
            evaluation_id: ID of the corresponding evaluation overview.
            evaluation_type: Type of evaluations that the :class:`Evaluator` returned
                in :func:`Evaluator.do_evaluate`.

        Returns:
            A :class:`Sequence` of successful :class:`ExampleEvaluation`s.
        """
        results = self.example_evaluations(evaluation_id, evaluation_type)
        return [r for r in results if not isinstance(r.result, FailedExampleEvaluation)]  # type: ignore

    def failed_example_evaluations(
        self, evaluation_id: str, evaluation_type: type[Evaluation]
    ) -> Sequence[ExampleEvaluation[FailedExampleEvaluation]]:
        """Returns all failed :class:`ExampleEvaluation`s for the given evaluation overview ID sorted by their example ID.

        Args:
            evaluation_id: ID of the corresponding evaluation overview.
            evaluation_type: Type of evaluations that the :class:`Evaluator` returned
                in :func:`Evaluator.do_evaluate`.

        Returns:
            A :class:`Sequence` of failed :class:`ExampleEvaluation`s.
        """
        results = self.example_evaluations(evaluation_id, evaluation_type)
        return [r for r in results if isinstance(r.result, FailedExampleEvaluation)]  # type: ignore



from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from typing import (
    Any,
)

from pydantic import BaseModel
from pydantic import Field as PydanticField


class ArgillaEvaluation(BaseModel):
    """The evaluation result for a single rating record in an Argilla feedback-dataset.

    Attributes:
        example_id: the id of the example that was evaluated.
        record_id: the id of the record that is evaluated.
        responses: Maps question-names (:attr:`Question.name` ) to response values.
        metadata: Metadata belonging to the evaluation, for example ids of completions.
    """

    example_id: str
    record_id: str
    responses: Mapping[str, Any]
    metadata: Mapping[str, Any]


class RecordData(BaseModel):
    """Input-data for a Argilla evaluation record.

    This can be used to add a new record to an existing Argilla feedback-dataset.
    Once it is added it gets an Argilla provided id and can be retrieved as :class:`Record`

    Attributes:
        content: Maps field-names (:attr:`Field.name` ) to string values that can be displayed to the user.
        example_id: the id of the corresponding :class:`Example` from a :class:`Dataset`.
        metadata: Arbitrary metadata in form of key/value strings that can be attached to a record.
    """

    content: Mapping[str, str]
    example_id: str
    metadata: Mapping[str, str | int] = PydanticField(default_factory=dict)


class Record(RecordData):
    """Represents an Argilla record of an feedback-dataset.

    Just adds the id to a :class:`RecordData`

    Attributes:
        id: the Argilla generated id of the record.
    """

    id: str


class ArgillaClient(ABC):
    """Client interface for accessing an Argilla server.

    Argilla supports human in the loop evaluation. This class defines the API used by
    the intelligence layer to create feedback datasets or retrieve evaluation results.
    """

    @abstractmethod
    def create_dataset(
        self,
        workspace_id: str,
        dataset_name: str,
        fields: Sequence[Any],
        questions: Sequence[Any],
    ) -> str:
        """Creates and publishes a new feedback dataset in Argilla.

        Raises an error if the name exists already.

        Args:
            workspace_id: the id of the workspace the feedback dataset should be created in.
                The user executing this request must have corresponding permissions for this workspace.
            dataset_name: the name of the feedback-dataset to be created.
            fields: all fields of this dataset.
            questions: all questions for this dataset.

        Returns:
            The id of the created dataset.
        """
        ...

    @abstractmethod
    def ensure_dataset_exists(
        self,
        workspace_id: str,
        dataset_name: str,
        fields: Sequence[Any],
        questions: Sequence[Any],
    ) -> str:
        """Retrieves an existing dataset or creates and publishes a new feedback dataset in Argilla.

        Args:
            workspace_id: the id of the workspace the feedback dataset should be created in.
                The user executing this request must have corresponding permissions for this workspace.
            dataset_name: the name of the feedback-dataset to be created.
            fields: all fields of this dataset.
            questions: all questions for this dataset.

        Returns:
            The id of the dataset to be retrieved .
        """
        ...

    @abstractmethod
    def add_record(self, dataset_id: str, record: RecordData) -> None:
        """Adds a new record to the given dataset.

        Args:
            dataset_id: id of the dataset the record is added to
            record: the actual record data (i.e. content for the dataset's fields)
        """
        ...

    def add_records(self, dataset_id: str, records: Sequence[RecordData]) -> None:
        """Adds new records to the given dataset.

        Args:
            dataset_id: id of the dataset the record is added to
            records: list containing the record data (i.e. content for the dataset's fields)
        """
        for record in records:
            return self.add_record(dataset_id, record)

    @abstractmethod
    def evaluations(self, dataset_id: str) -> Iterable[ArgillaEvaluation]:
        """Returns all human-evaluated evaluations for the given dataset.

        Args:
            dataset_id: the id of the dataset.

        Returns:
            An `Iterable` over all human-evaluated evaluations for the given dataset.
        """
        ...

    @abstractmethod
    def split_dataset(self, dataset_id: str, n_splits: int) -> None:
        """Adds a new metadata property to the dataset and assigns a split to each record.

        Args:
            dataset_id: the id of the dataset
            n_splits: the number of splits to create
        """
        ...



import contextlib
from collections.abc import Iterable
from pathlib import Path
from typing import Optional

from fsspec.implementations.local import LocalFileSystem  # type: ignore
from pharia_inference_sdk.core import Input, JsonSerializer, PydanticSerializable

from pharia_studio_sdk.connectors.base.json_serializable import SerializableDict
from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import (
    Dataset,
    Example,
    ExpectedOutput,
)
from pharia_studio_sdk.evaluation.infrastructure.file_system_based_repository import (
    FileSystemBasedRepository,
)


class FileSystemDatasetRepository(DatasetRepository, FileSystemBasedRepository):
    """A dataset repository that stores :class:`Dataset`s in files.

    It creates a single file per dataset and stores the :class:`Example`s as lines in this file.
    The format of the file is `.jsonl`.
    """

    _REPO_TYPE = "dataset"

    def create_dataset(
        self,
        examples: Iterable[Example[Input, ExpectedOutput]],
        dataset_name: str,
        id: str | None = None,
        labels: set[str] | None = None,
        metadata: SerializableDict | None = None,
    ) -> Dataset:
        if metadata is None:
            metadata = dict()
        if labels is None:
            labels = set()
        dataset = Dataset(name=dataset_name, labels=labels, metadata=metadata)
        if id is not None:
            dataset.id = id

        self.mkdir(self._dataset_directory(dataset.id))

        dataset_path = self._dataset_path(dataset.id)
        examples_path = self._dataset_examples_path(dataset.id)
        if self.exists(dataset_path) or self.exists(examples_path):
            raise ValueError(
                f"One of the dataset files already exist for dataset {dataset}. This should not happen. Files: {dataset_path}, {examples_path}."
            )

        self._write_data(dataset_path, [dataset])
        self._write_data(examples_path, examples)

        return dataset

    def delete_dataset(self, dataset_id: str) -> None:
        with contextlib.suppress(FileNotFoundError):
            self._file_system.rm(
                self.path_to_str(self._dataset_directory(dataset_id)), recursive=True
            )

    def dataset(self, dataset_id: str) -> Optional[Dataset]:
        file_path = self.path_to_str(self._dataset_path(dataset_id))
        if not self._file_system.exists(file_path):
            return None

        with self._file_system.open(file_path, "r", encoding="utf-8") as file_content:
            # we save only one dataset per file
            return [
                Dataset.model_validate_json(dataset_string)
                for dataset_string in file_content
            ][0]

    def dataset_ids(self) -> Iterable[str]:
        dataset_files = self._file_system.glob(
            path=self.path_to_str(self._dataset_root_directory()) + "/**/*.jsonl",
            maxdepth=2,
            detail=False,
        )
        return sorted([Path(f).stem for f in dataset_files])

    def example(
        self,
        dataset_id: str,
        example_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
    ) -> Optional[Example[Input, ExpectedOutput]]:
        example_path = self._dataset_examples_path(dataset_id)
        if not self.exists(example_path.parent):
            raise ValueError(
                f"Repository does not contain a dataset with id: {dataset_id}"
            )
        if not self.exists(example_path):
            return None
        for example in self.examples(dataset_id, input_type, expected_output_type):
            if example.id == example_id:
                return example
        return None

    def examples(
        self,
        dataset_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
        examples_to_skip: Optional[frozenset[str]] = None,
    ) -> Iterable[Example[Input, ExpectedOutput]]:
        examples_to_skip = examples_to_skip or frozenset()
        example_path = self.path_to_str(self._dataset_examples_path(dataset_id))
        if not self._file_system.exists(example_path):
            raise ValueError(
                f"Repository does not contain a dataset with id: {dataset_id}"
            )

        with self._file_system.open(
            example_path, "r", encoding="utf-8"
        ) as examples_file:
            # Mypy does not accept dynamic types
            examples = []
            for example in examples_file:
                current_example = Example[
                    input_type, expected_output_type  # type: ignore
                ].model_validate_json(json_data=example)
                if current_example.id not in examples_to_skip:
                    examples.append(current_example)

        return sorted(examples, key=lambda example: example.id)

    def _dataset_root_directory(self) -> Path:
        return self._root_directory / "datasets"

    def _dataset_directory(self, dataset_id: str) -> Path:
        return self._dataset_root_directory() / f"{dataset_id}"

    def _dataset_path(self, dataset_id: str) -> Path:
        return self._dataset_directory(dataset_id) / f"{dataset_id}.json"

    def _dataset_examples_path(self, dataset_id: str) -> Path:
        return self._dataset_directory(dataset_id) / f"{dataset_id}.jsonl"

    def _write_data(
        self,
        file_path: Path,
        data_to_write: Iterable[PydanticSerializable],
    ) -> None:
        data = "\n".join(
            JsonSerializer(root=chunk).model_dump_json() for chunk in data_to_write
        )
        self.write_utf8(file_path, data, create_parents=True)


class FileDatasetRepository(FileSystemDatasetRepository):
    def __init__(self, root_directory: Path) -> None:
        super().__init__(LocalFileSystem(), root_directory)



from collections.abc import Iterable
from typing import Optional

from pharia_inference_sdk.core import Input

from pharia_studio_sdk.connectors import (
    SerializableDict,
    StudioClient,
)
from pharia_studio_sdk.connectors.studio.studio import (
    StudioDataset,
    StudioExample,
)
from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import (
    Dataset,
    Example,
    ExpectedOutput,
)
from pharia_studio_sdk.evaluation.dataset.in_memory_dataset_repository import (
    InMemoryDatasetRepository,
)


class StudioDatasetRepository(DatasetRepository):
    """Dataset repository interface with Data Platform."""

    def __init__(self, studio_client: StudioClient) -> None:
        """Initializes the StudioDatasetRepository.

        Args:
            studio_client: Client to interact with the Studio API.
        """
        self.studio_client = studio_client
        self._in_memory_dataset_repository = InMemoryDatasetRepository()

    def create_dataset(
        self,
        examples: Iterable[Example[Input, ExpectedOutput]],
        dataset_name: str,
        id: str | None = None,
        labels: set[str] | None = None,
        metadata: SerializableDict | None = None,
    ) -> Dataset:
        """Creates a dataset from given :class:`Example`s and returns the ID of that dataset.

        Args:
            examples: An :class:`Iterable` of :class:`Example`s to be saved in the same dataset.
            dataset_name: A name for the dataset.
            id: ID is not used in the StudioDatasetRepository as it is generated by the Studio.
            labels: A list of labels for filtering. Defaults to an empty list. Defaults to None.
            metadata: A dict for additional information about the dataset. Defaults to an empty dict. Defaults to None.

        Returns:
           :class:`Dataset`
        """
        if id is not None:
            raise NotImplementedError(
                "Custom dataset IDs are not supported by the StudioDataRepository"
            )

        created_dataset = Dataset(
            name=dataset_name,
            labels=labels or set(),
            metadata=metadata or dict(),
        )

        studio_dataset = self.map_to_studio_dataset(created_dataset)
        studio_examples = self.map_to_many_studio_example(examples)

        studio_dataset_id = self.studio_client.submit_dataset(
            dataset=studio_dataset, examples=studio_examples
        )

        created_dataset.id = studio_dataset_id
        return created_dataset

    def delete_dataset(self, dataset_id: str) -> None:
        """Deletes a dataset identified by the given dataset ID.

        Args:
            dataset_id: Dataset ID of the dataset to delete.

        """
        raise NotImplementedError()

    def dataset(self, dataset_id: str) -> Optional[Dataset]:
        """Returns a dataset identified by the given dataset ID.

        Args:
            dataset_id: Dataset ID of the dataset to delete.

        Returns:
            :class:`Dataset` if it was not, `None` otherwise.
        """
        raise NotImplementedError()

    def datasets(self) -> Iterable[Dataset]:
        """Returns all :class:`Dataset`s. Sorting is not guaranteed.

        Returns:
            :class:`Sequence` of :class:`Dataset`s.
        """
        raise NotImplementedError()

    def dataset_ids(self) -> Iterable[str]:
        """Returns all sorted dataset IDs.

        Returns:
            :class:`Iterable` of dataset IDs.
        """
        raise NotImplementedError()

    def example(
        self,
        dataset_id: str,
        example_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
    ) -> Optional[Example[Input, ExpectedOutput]]:
        """Returns an :class:`Example` for the given dataset ID and example ID.

        Args:
            dataset_id: Dataset ID of the linked dataset.
            example_id: ID of the example to retrieve.
            input_type: Input type of the example.
            expected_output_type: Expected output type of the example.

        Returns:
            :class:`Example` if it was found, `None` otherwise.
        """
        raise NotImplementedError()

    def examples(
        self,
        dataset_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
        examples_to_skip: Optional[frozenset[str]] = None,
    ) -> Iterable[Example[Input, ExpectedOutput]]:
        """Returns all :class:`Example`s for the given dataset ID sorted by their ID.

        Args:
            dataset_id: Dataset ID whose examples should be retrieved.
            input_type: Input type of the example.
            expected_output_type: Expected output type of the example.
            examples_to_skip: Optional list of example IDs. Those examples will be excluded from the output. Defaults to None.

        Returns:
            :class:`Iterable` of :class`Example`s.
        """
        if self._in_memory_dataset_repository.dataset(dataset_id) is None:
            examples = self.map_to_many_example(
                self.studio_client.get_dataset_examples(
                    dataset_id,
                    input_type=input_type,
                    expected_output_type=expected_output_type,
                )
            )
            self._in_memory_dataset_repository.create_dataset(
                examples, id=dataset_id, dataset_name="in_memory_dataset"
            )

        return self._in_memory_dataset_repository.examples(
            dataset_id,
            input_type,
            expected_output_type,
            examples_to_skip=examples_to_skip,
        )

    @staticmethod
    def map_to_studio_example(
        example_to_map: Example[Input, ExpectedOutput],
    ) -> StudioExample[Input, ExpectedOutput]:
        # (**.model_dump) drops typing information, so we manually assign
        return StudioExample(
            input=example_to_map.input,
            expected_output=example_to_map.expected_output,
            id=example_to_map.id,
            metadata=example_to_map.metadata,
        )

    @staticmethod
    def map_to_many_studio_example(
        examples_to_map: Iterable[Example[Input, ExpectedOutput]],
    ) -> Iterable[StudioExample[Input, ExpectedOutput]]:
        return (
            StudioDatasetRepository.map_to_studio_example(example)
            for example in examples_to_map
        )

    @staticmethod
    def map_to_studio_dataset(dataset_to_map: Dataset) -> StudioDataset:
        return StudioDataset(**dataset_to_map.model_dump())

    @staticmethod
    def map_to_example(
        example_to_map: StudioExample[Input, ExpectedOutput],
    ) -> Example[Input, ExpectedOutput]:
        # (**.model_dump) drops typing information, so we manually assign
        return Example[Input, ExpectedOutput](
            input=example_to_map.input,
            expected_output=example_to_map.expected_output,
            id=example_to_map.id,
            metadata=example_to_map.metadata,
        )

    @staticmethod
    def map_to_many_example(
        examples_to_map: Iterable[StudioExample[Input, ExpectedOutput]],
    ) -> Iterable[Example[Input, ExpectedOutput]]:
        return (
            StudioDatasetRepository.map_to_example(example)
            for example in examples_to_map
        )



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")
Output = TypeVar("Output")


class Accumulator(ABC, Generic[T, Output]):
    """Used for incremental computation.

    For use cases with large amount of data where you don't want to have every value in memory at once, e.g. evaluation.
    """

    @abstractmethod
    def add(self, value: T) -> None:
        """Responsible for accumulating values.

        Args:
            value: the value to add
        Returns:
             nothing
        """
        ...

    @abstractmethod
    def extract(self) -> Output:
        """Accumulates the final result.

        Returns:
           float: 0.0 if no values were added before, else the mean
        """
        ...


class MeanAccumulator(Accumulator[float, float]):
    def __init__(self) -> None:
        self._n = 0
        self._acc = 0.0
        self._squares_acc = 0.0  # Sum of squares of the values

    def add(self, value: float) -> None:
        self._n += 1
        self._acc += value
        self._squares_acc += value**2

    def extract(self) -> float:
        """Accumulates the mean.

        :return: 0.0 if no values were added before, else the mean
        """
        return 0.0 if self._n == 0 else self._acc / self._n

    def standard_deviation(self) -> float:
        """Calculates the standard deviation."""
        if self._n == 0:
            return 0.0
        mean = self.extract()
        variance = (self._squares_acc / self._n) - (mean**2)
        return variance**0.5

    def standard_error(self) -> float:
        """Calculates the standard error of the mean."""
        if self._n <= 1:
            return 0.0
        return self.standard_deviation() / (self._n**0.5)



import warnings
from collections import defaultdict
from collections.abc import Iterable, Sequence
from typing import Any, Optional, cast

from pharia_inference_sdk.core import InMemoryTracer, Output, PydanticSerializable
from pharia_inference_sdk.core.tracer.tracer import Tracer
from pydantic import TypeAdapter

from pharia_studio_sdk.evaluation.run.domain import (
    ExampleOutput,
    FailedExampleRun,
    RunOverview,
)
from pharia_studio_sdk.evaluation.run.run_repository import RecoveryData, RunRepository


class InMemoryRunRepository(RunRepository):
    def __init__(self) -> None:
        super().__init__()
        self._example_outputs: dict[str, list[ExampleOutput[PydanticSerializable]]] = (
            defaultdict(list)
        )
        self._example_traces: dict[str, Tracer] = dict()
        self._run_overviews: dict[str, RunOverview] = dict()
        self._recovery_data: dict[str, RecoveryData] = dict()

    def store_run_overview(self, overview: RunOverview) -> None:
        self._run_overviews[overview.id] = overview
        if overview.id not in self._example_outputs:
            self._example_outputs[overview.id] = []

    def _create_temporary_run_data(self, tmp_hash: str, run_id: str) -> None:
        self._recovery_data[tmp_hash] = RecoveryData(run_id=run_id)

    def _delete_temporary_run_data(self, tmp_hash: str) -> None:
        del self._recovery_data[tmp_hash]

    def _temp_store_finished_example(self, tmp_hash: str, example_id: str) -> None:
        self._recovery_data[tmp_hash].finished_examples.append(example_id)

    def finished_examples(self, tmp_hash: str) -> Optional[RecoveryData]:
        if tmp_hash in self._recovery_data:
            return self._recovery_data[tmp_hash]
        else:
            return None

    def run_overview(self, run_id: str) -> Optional[RunOverview]:
        return self._run_overviews.get(run_id, None)

    def run_overview_ids(self) -> Sequence[str]:
        return sorted(self._run_overviews.keys())

    def store_example_output(self, example_output: ExampleOutput[Output]) -> None:
        self._example_outputs[example_output.run_id].append(
            cast(ExampleOutput[PydanticSerializable], example_output)
        )

    @staticmethod
    def _convert_to_type(data: Any, desired_type: type[Output]) -> Output:
        if type(data) is desired_type:
            return data
        else:
            return TypeAdapter(desired_type).validate_python(data)

    def _generate_output_from_internal_output(
        self,
        internal_output: ExampleOutput[PydanticSerializable],
        output_type: type[Output],
    ) -> ExampleOutput[Output] | ExampleOutput[FailedExampleRun]:
        if (
            type(internal_output.output) is output_type
            or type(internal_output.output) is FailedExampleRun
        ):
            return internal_output  # type: ignore
        return ExampleOutput[Output](
            run_id=internal_output.run_id,
            example_id=internal_output.example_id,
            output=self._convert_to_type(internal_output.output, output_type),
        )

    def example_output(
        self, run_id: str, example_id: str, output_type: type[Output]
    ) -> Optional[ExampleOutput[Output] | ExampleOutput[FailedExampleRun]]:
        if run_id not in self._example_outputs:
            warnings.warn(
                f'Repository does not contain a run with id: "{run_id}"', UserWarning
            )
            return None

        for example_output in self._example_outputs[run_id]:
            if example_output.example_id == example_id:
                return self._generate_output_from_internal_output(
                    example_output, output_type
                )
        return None

    def example_outputs(
        self, run_id: str, output_type: type[Output]
    ) -> Iterable[ExampleOutput[Output] | ExampleOutput[FailedExampleRun]]:
        if run_id not in self._example_outputs and run_id not in self._run_overviews:
            warnings.warn(
                f'Repository does not contain a run with id: "{run_id}"', UserWarning
            )
            return []

        return (
            self._generate_output_from_internal_output(example_output, output_type)
            for example_output in sorted(
                self._example_outputs[run_id],
                key=lambda example_output: example_output.example_id,
            )
        )

    def example_output_ids(self, run_id: str) -> Sequence[str]:
        return sorted(
            [
                example_output.example_id
                for example_output in self._example_outputs[run_id]
            ]
        )

    def example_tracer(self, run_id: str, example_id: str) -> Optional[Tracer]:
        return self._example_traces.get(f"{run_id}/{example_id}")

    def create_tracer_for_example(self, run_id: str, example_id: str) -> Tracer:
        tracer = InMemoryTracer()
        self._example_traces[f"{run_id}/{example_id}"] = tracer
        return tracer



from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Optional

from pharia_inference_sdk.core import Input

from pharia_studio_sdk.connectors.base.json_serializable import (
    SerializableDict,
)
from pharia_studio_sdk.evaluation.dataset.domain import (
    Dataset,
    Example,
    ExpectedOutput,
)


class DatasetRepository(ABC):
    """Base dataset repository interface.

    Provides methods to store and load datasets and their linked examples (:class:`Example`s).
    """

    @abstractmethod
    def create_dataset(
        self,
        examples: Iterable[Example[Input, ExpectedOutput]],
        dataset_name: str,
        id: str | None = None,
        labels: set[str] | None = None,
        metadata: SerializableDict | None = None,
    ) -> Dataset:
        """Creates a dataset from given :class:`Example`s and returns the ID of that dataset.

        Args:
            examples: An :class:`Iterable` of :class:`Example`s to be saved in the same dataset.
            dataset_name: A name for the dataset.
            id: The dataset ID. If `None`, an ID will be generated.
            labels: A list of labels for filtering. Defaults to an empty list.
            metadata: A dict for additional information about the dataset. Defaults to an empty dict.

        Returns:
            The created :class:`Dataset`.
        """
        pass

    @abstractmethod
    def delete_dataset(self, dataset_id: str) -> None:
        """Deletes a dataset identified by the given dataset ID.

        Args:
            dataset_id: Dataset ID of the dataset to delete.
        """
        pass

    @abstractmethod
    def dataset(self, dataset_id: str) -> Optional[Dataset]:
        """Returns a dataset identified by the given dataset ID.

        Args:
            dataset_id: Dataset ID of the dataset to delete.

        Returns:
            :class:`Dataset` if it was not, `None` otherwise.
        """
        pass

    def datasets(self) -> Iterable[Dataset]:
        """Returns all :class:`Dataset`s sorted by their ID.

        Yields:
            :class:`Dataset`s.
        """
        for dataset_id in self.dataset_ids():
            dataset = self.dataset(dataset_id)
            if dataset is not None:
                yield dataset

    @abstractmethod
    def dataset_ids(self) -> Iterable[str]:
        """Returns all sorted dataset IDs.

        Returns:
            :class:`Iterable` of dataset IDs.
        """
        pass

    @abstractmethod
    def example(
        self,
        dataset_id: str,
        example_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
    ) -> Optional[Example[Input, ExpectedOutput]]:
        """Returns an :class:`Example` for the given dataset ID and example ID.

        Args:
            dataset_id: Dataset ID of the linked dataset.
            example_id: ID of the example to retrieve.
            input_type: Input type of the example.
            expected_output_type: Expected output type of the example.

        Returns:
            :class:`Example` if it was found, `None` otherwise.
        """
        pass

    @abstractmethod
    def examples(
        self,
        dataset_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
        examples_to_skip: Optional[frozenset[str]] = None,
    ) -> Iterable[Example[Input, ExpectedOutput]]:
        """Returns all :class:`Example`s for the given dataset ID sorted by their ID.

        Args:
            dataset_id: Dataset ID whose examples should be retrieved.
            input_type: Input type of the example.
            expected_output_type: Expected output type of the example.
            examples_to_skip: Optional list of example IDs. Those examples will be excluded from the output.

        Returns:
            :class:`Iterable` of :class`Example`s.
        """
        pass



import random
from collections import Counter, defaultdict
from collections.abc import Iterable, Mapping, Sequence

import numpy as np
from pydantic import BaseModel

from pharia_studio_sdk.evaluation.aggregation.accumulator import MeanAccumulator
from pharia_studio_sdk.evaluation.aggregation.aggregator import AggregationLogic
from pharia_studio_sdk.evaluation.evaluation.evaluator.incremental_evaluator import (
    ComparisonEvaluation,
    Matches,
    MatchOutcome,
)


class PlayerScore(BaseModel):
    elo: float
    elo_standard_error: float
    win_rate: float
    num_matches: int


class AggregatedComparison(BaseModel):
    scores: Mapping[str, PlayerScore]


class EloAggregationAdapter:
    @staticmethod
    def aggregate(evaluations: Iterable[ComparisonEvaluation]) -> AggregatedComparison:
        evaluations = list(evaluations)
        player_counter = Counter(
            player
            for comparison_evaluation in evaluations
            for player in [
                comparison_evaluation.first_player,
                comparison_evaluation.second_player,
            ]
        )

        player_counts = dict(player_counter)
        players = player_counts.keys()

        accumulators = {p: MeanAccumulator() for p in players}
        for _ in range(100):
            elo_calc = EloCalculator(players)
            random.shuffle(evaluations)
            elo_calc.calculate(evaluations)
            for p in players:
                accumulators[p].add(elo_calc.ratings[p])

        win_rate_calc = WinRateCalculator(players)
        win_rate = win_rate_calc.calculate(evaluations)

        return AggregatedComparison(
            scores={
                p: PlayerScore(
                    elo=acc.extract(),
                    elo_standard_error=acc.standard_error(),
                    win_rate=win_rate[p],
                    num_matches=player_counts[p],
                )
                for p, acc in accumulators.items()
            },
        )


class EloCalculator:
    def __init__(
        self,
        players: Iterable[str],
        k_start: float = 20.0,
        k_floor: float = 10.0,
        decay_factor: float = 0.0005,
    ) -> None:
        self.ratings: dict[str, float] = {player: 1500.0 for player in players}
        self._match_counts: dict[str, int] = defaultdict(int)
        self._k_ceiling = k_start - k_floor
        self._k_floor = k_floor
        self._decay_factor = decay_factor

    def _calc_k_factor(self, player: str) -> float:
        n = self._match_counts.get(player) or 0
        return self._k_ceiling * np.exp(-self._decay_factor * n) + self._k_floor

    def _calc_expected_win_rates(
        self, player_a: str, player_b: str
    ) -> tuple[float, float]:
        rating_a, rating_b = self.ratings[player_a], self.ratings[player_b]
        exp_a = 1 / (1 + 10 ** ((rating_b - rating_a) / 400))
        return exp_a, 1 - exp_a

    def _calc_difs(
        self, match_outcome: MatchOutcome, player_a: str, player_b: str
    ) -> tuple[float, float]:
        expected_win_rate_a, expected_win_rate_b = self._calc_expected_win_rates(
            player_a, player_b
        )
        actual_a, actual_b = match_outcome.payoff
        k_a, k_b = self._calc_k_factor(player_a), self._calc_k_factor(player_b)
        return k_a * (actual_a - expected_win_rate_a), k_b * (
            actual_b - expected_win_rate_b
        )

    def calculate(self, matches: Sequence[ComparisonEvaluation]) -> None:
        for match in matches:
            dif_a, dif_b = self._calc_difs(
                match.outcome, match.first_player, match.second_player
            )
            self.ratings[match.first_player] += dif_a
            self.ratings[match.second_player] += dif_b
            self._match_counts[match.first_player] += 1
            self._match_counts[match.second_player] += 1


class WinRateCalculator:
    def __init__(self, players: Iterable[str]) -> None:
        self.match_count: dict[str, int] = {p: 0 for p in players}
        self.win_count: dict[str, float] = {p: 0 for p in players}

    def calculate(self, matches: Sequence[ComparisonEvaluation]) -> Mapping[str, float]:
        for match in matches:
            self.match_count[match.first_player] += 1
            self.match_count[match.second_player] += 1
            self.win_count[match.first_player] += match.outcome.payoff[0]
            self.win_count[match.second_player] += match.outcome.payoff[1]

        return {
            player: self.win_count[player] / match_count
            for player, match_count in self.match_count.items()
        }


class ComparisonEvaluationAggregationLogic(
    AggregationLogic[ComparisonEvaluation, AggregatedComparison]
):
    def aggregate(
        self, evaluations: Iterable[ComparisonEvaluation]
    ) -> AggregatedComparison:
        return EloAggregationAdapter.aggregate(evaluations)


class MatchesAggregationLogic(AggregationLogic[Matches, AggregatedComparison]):
    def aggregate(self, evaluations: Iterable[Matches]) -> AggregatedComparison:
        flattened_matches = [
            comparison_evaluation
            for match in evaluations
            for comparison_evaluation in match.comparison_evaluations
        ]
        return EloAggregationAdapter.aggregate(flattened_matches)



from collections.abc import Iterable, Sequence
from datetime import datetime
from typing import Generic, TypeVar

from pydantic import BaseModel, SerializeAsAny

from pharia_studio_sdk.connectors.base.json_serializable import (
    SerializableDict,
)
from pharia_studio_sdk.evaluation.evaluation.domain import (
    EvaluationFailed,
    EvaluationOverview,
)
from pharia_studio_sdk.evaluation.run.domain import RunOverview

AggregatedEvaluation = TypeVar("AggregatedEvaluation", bound=BaseModel, covariant=True)


class AggregationOverview(BaseModel, Generic[AggregatedEvaluation], frozen=True):
    """Complete overview of the results of evaluating a :class:`Task` on a dataset.

    Created when running :meth:`Evaluator.eval_and_aggregate_runs`. Contains high-level information and statistics.

    Attributes:
        evaluation_overviews: :class:`EvaluationOverview`s used for aggregation.
        id: Aggregation overview ID.
        start: Start timestamp of the aggregation.
        end: End timestamp of the aggregation.
        end: The time when the evaluation run ended
        successful_evaluation_count: The number of examples that where successfully evaluated.
        crashed_during_evaluation_count: The number of examples that crashed during evaluation.
        failed_evaluation_count: The number of examples that crashed during evaluation
            plus the number of examples that failed to produce an output for evaluation.
        run_ids: IDs of all :class:`RunOverview`s from all linked :class:`EvaluationOverview`s.
        description: A short description.
        statistics: Aggregated statistics of the run. Whatever is returned by :meth:`Evaluator.aggregate`
        labels: Labels for filtering aggregation. Defaults to empty list.
        metadata: Additional information about the aggregation. Defaults to empty dict.

    """

    evaluation_overviews: frozenset[EvaluationOverview]
    id: str
    start: datetime
    end: datetime
    successful_evaluation_count: int
    crashed_during_evaluation_count: int
    description: str
    statistics: SerializeAsAny[AggregatedEvaluation]
    labels: set[str] = set()
    metadata: SerializableDict = dict()

    @property
    def run_ids(self) -> Sequence[str]:
        return [overview.id for overview in self.run_overviews()]

    def run_overviews(self) -> Iterable[RunOverview]:
        return set(
            run_overview
            for evaluation_overview in self.evaluation_overviews
            for run_overview in evaluation_overview.run_overviews
        )

    @property
    def failed_evaluation_count(self) -> int:
        return self.crashed_during_evaluation_count + sum(
            run_overview.failed_example_count for run_overview in self.run_overviews()
        )

    def raise_on_evaluation_failure(self) -> None:
        if self.crashed_during_evaluation_count > 0:
            raise EvaluationFailed(self.id, self.crashed_during_evaluation_count)

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        res = (
            f"Aggregation Overview ID = {self.id}\n"
            f"Start time = {self.start}\n"
            f"End time = {self.end}\n"
            f"Successful example count = {self.successful_evaluation_count}\n"
            f"Count of examples crashed during evaluation = {self.failed_evaluation_count}\n"
            f'Description = "{self.description}"\n'
            f"Labels = {self.labels}\n"
            f"Metadata = {self.metadata}\n"
        )

        res += f"IDs of aggregated Evaluation Overviews = {[evaluation_overview.id for evaluation_overview in self.evaluation_overviews]}\n"
        res += f"IDs of aggregated Run Overviews = {self.run_ids}\n"

        res += "Statistics = {\n"
        res += f"{self.statistics}\n"
        res += "}\n"

        return res

    def __hash__(self) -> int:
        return hash(self.id)



import math
from collections.abc import Sequence
from dataclasses import dataclass

from lingua import LanguageDetectorBuilder
from rouge_score import rouge_scorer
from sacrebleu import BLEU
from semantic_text_splitter import TextSplitter


class BleuGrader:
    def __init__(self) -> None:
        self.bleu = BLEU()

    def calculate_bleu(self, hypothesis: str, reference: str) -> float:
        """Calculates the BLEU-score for the given hypothesis and reference.

        In the summarization use-case the `BLEU-score <https://aclanthology.org/P02-1040/>`_ roughly corresponds to the precision of the generated summary with regard to the expected summary.

        Args:
            hypothesis: The generation to be evaluated.
            reference: The baseline for the evaluation.

        Returns:
            BLEU-score, float between 0 and 100. Where 100 means perfect match and 0 no overlap.
        """
        bleu_score = self.bleu.corpus_score(
            hypotheses=[hypothesis], references=[[reference]]
        )

        return bleu_score.score


@dataclass
class FScores:
    precision: float
    recall: float
    f_score: float


class RougeGrader:
    def __init__(self) -> None:
        self.scorer = rouge_scorer.RougeScorer(["rouge2"], use_stemmer=True)

    def calculate_rouge(self, hypothesis: str, reference: str) -> FScores:
        """Calculates the ROUGE-score for the hypothesis and reference.

        In the summarization use-case the `ROUGE-score <https://aclanthology.org/W04-1013>`_ roughly corresponds to the recall of the generated summary with regard to the expected summary.

        Args:
            hypothesis: The generation to be evaluated.
            reference: The baseline for the evaluation.

        Returns:
            ROUGE-score, which contains precision, recall and f1 metrics, all will be floats between 0 and 1. Where 1 means perfect match and 0 no overlap.
        """
        rouge_score = self.scorer.score(prediction=hypothesis, target=reference)[
            "rouge2"
        ]
        return FScores(
            precision=rouge_score.precision,
            recall=rouge_score.recall,
            f_score=rouge_score.fmeasure,
        )


class LanguageMatchesGrader:
    """Provides a method to evaluate whether two texts are of the same language.

    Args:
        acceptance_threshold: probability a language must surpass to be accepted
    """

    _acceptance_threshold: float

    def __init__(self, acceptance_threshold: float = 0.1) -> None:
        self._acceptance_threshold = acceptance_threshold
        self._detector = (
            LanguageDetectorBuilder.from_all_languages_with_latin_script().build()
        )

    def languages_match(self, input: str, output: str) -> bool:
        """Calculates if the input and output text are of the same language.

        The length of the texts and its sentences should be reasonably long in order for good performance.

        Args:
            input: text for which languages is compared to
            output: text

        Returns:
            bool: whether input and output language match
                  returns true if clear input language is not determinable
        """
        dominant_input_language = self._get_dominant_language(input)

        if dominant_input_language is None:
            return True

        dominant_output_language = self._get_dominant_language(output)

        return dominant_input_language == dominant_output_language

    def _get_dominant_language(self, text: str) -> str | None:
        test_chunks: Sequence[str] = self._tokenize_text(text)
        probs_per_language = self._get_scores_per_language(test_chunks)
        languages_above_threshold = [
            langs
            for langs, probs in probs_per_language.items()
            if probs >= self._acceptance_threshold
        ]

        if len(languages_above_threshold) != 1:
            return None

        return languages_above_threshold[0]

    @staticmethod
    def _tokenize_text(
        text: str, lower_char_bound: int = 30, upper_char_bound: int = 200
    ) -> Sequence[str]:
        return TextSplitter((lower_char_bound, upper_char_bound)).chunks(text)

    def _get_scores_per_language(self, text_chunks: Sequence[str]) -> dict[str, float]:
        scores_per_language: dict[str, float] = {}
        for text_chunk in text_chunks:
            languages_with_probs = self._detector.compute_language_confidence_values(
                text_chunk
            )
            for lang in languages_with_probs:
                scores_per_language[lang.language.name] = scores_per_language.get(
                    lang.language.name, 0
                ) + lang.value * len(text_chunk)

        return self._normalize_dict(scores_per_language)

    @staticmethod
    def _normalize_dict(dictionary: dict[str, float]) -> dict[str, float]:
        total = sum(dictionary.values())
        if total == 0:
            return {key: 0 for key in dictionary}
        return {key: value / total for key, value in dictionary.items()}


@dataclass
class IndexRange:
    start: int
    stop: int


_HighlightRange = list[IndexRange]


class HighlightCoverageGrader:
    """Evaluates how well the generated highlights match the expected highlights (via precision, recall and f1-score).

    Args:
        beta_factor: factor to control weight of precision (0 <= beta < 1) vs. recall (beta > 1) when computing the f-score
    """

    beta_factor: float

    def __init__(self, beta_factor: float = 1.0) -> None:
        self.beta_factor = beta_factor

    def compute_fscores(
        self,
        generated_highlight_indices: Sequence[tuple[int, int]],
        expected_highlight_indices: Sequence[tuple[int, int]],
    ) -> FScores:
        """Calculates how well the generated highlight ranges match the expected ones.

        Args:
            generated_highlight_indices: list of tuples(start, end) of the generated highlights
            expected_highlight_indices: list of tuples(start, end) of the generated highlights

        Returns:
            FScores, which contains precision, recall and f-score metrics, all will be floats between 0 and 1,
            where 1 means perfect match and 0 no overlap
        """
        generated_highlight_ranges: _HighlightRange = [
            IndexRange(el[0], el[1]) for el in generated_highlight_indices
        ]
        expected_highlight_ranges: _HighlightRange = [
            IndexRange(el[0], el[1]) for el in expected_highlight_indices
        ]

        (
            correctly_identified_indices,
            false_positive_indices,
            false_negative_indices,
        ) = self._identify_overlap_ranges(
            generated_highlight_ranges, expected_highlight_ranges
        )

        true_positive_length = sum(
            [
                index_range.stop - index_range.start
                for index_range in correctly_identified_indices
            ]
        )
        false_positive_length = sum(
            [
                index_range.stop - index_range.start
                for index_range in false_positive_indices
            ]
        )
        false_negative_length = sum(
            [
                index_range.stop - index_range.start
                for index_range in false_negative_indices
            ]
        )

        precision = true_positive_length / (
            true_positive_length + false_positive_length
        )
        recall = true_positive_length / (true_positive_length + false_negative_length)

        denominator = math.pow(self.beta_factor, 2) * precision + recall
        if denominator == 0.0:
            f1_score = 0.0
        else:
            f1_score = (
                (1 + math.pow(self.beta_factor, 2)) * precision * recall
            ) / denominator

        return FScores(precision=precision, recall=recall, f_score=f1_score)

    @staticmethod
    def _identify_overlap_ranges(
        generated_highlights: _HighlightRange, expected_highlights: _HighlightRange
    ) -> tuple[_HighlightRange, _HighlightRange, _HighlightRange]:
        max_index: int = max(
            index_range.stop
            for index_range in generated_highlights + expected_highlights
        )

        def get_highlight_present_array(highlights: _HighlightRange) -> Sequence[bool]:
            highlight_map = [False] * max_index
            for index_range in highlights:
                for index in range(index_range.start, index_range.stop):
                    highlight_map[index] = True
            return highlight_map

        gen_highlight_present_array = get_highlight_present_array(generated_highlights)
        exp_highlight_present_array = get_highlight_present_array(expected_highlights)

        overlapping_indices: _HighlightRange = []
        leftover_gen_highlights: _HighlightRange = []
        leftover_exp_highlights: _HighlightRange = []
        current_range: IndexRange | None = None

        for index, (
            generated_highlight_present,
            expected_highlight_present,
        ) in enumerate(
            zip(gen_highlight_present_array, exp_highlight_present_array, strict=True)
        ):
            if generated_highlight_present and expected_highlight_present:
                current_range = HighlightCoverageGrader._increase_current_range_by_one(
                    current_range, index
                )
            else:
                if current_range:
                    overlapping_indices.append(current_range)
                    current_range = None

                if generated_highlight_present != expected_highlight_present:
                    if generated_highlight_present:
                        leftover_highlights = leftover_gen_highlights
                    else:
                        leftover_highlights = leftover_exp_highlights

                    HighlightCoverageGrader._increase_last_leftover_range_by_one(
                        index, leftover_highlights
                    )

        if current_range:
            overlapping_indices.append(current_range)

        return overlapping_indices, leftover_gen_highlights, leftover_exp_highlights

    @staticmethod
    def _increase_current_range_by_one(
        current_range: IndexRange | None, index: int
    ) -> IndexRange:
        if current_range is None:
            return IndexRange(index, index + 1)
        return IndexRange(current_range.start, index + 1)

    @staticmethod
    def _increase_last_leftover_range_by_one(
        index: int, leftover_highlights: _HighlightRange
    ) -> _HighlightRange:
        if leftover_highlights and leftover_highlights[-1].stop == index:
            leftover_highlights[-1] = (
                HighlightCoverageGrader._increase_current_range_by_one(
                    leftover_highlights[-1], index
                )
            )
        else:
            leftover_highlights.append(IndexRange(index, index + 1))

        return leftover_highlights



import traceback
from datetime import datetime
from typing import Generic, TypeVar

from pydantic import BaseModel, SerializeAsAny
from rich.tree import Tree

from pharia_studio_sdk.connectors.base.json_serializable import SerializableDict
from pharia_studio_sdk.evaluation.run.domain import RunOverview

Evaluation = TypeVar("Evaluation", bound=BaseModel, covariant=True)
"""Custom type that holds the domain-specific data of a single :class:`Example` evaluation"""


class FailedExampleEvaluation(BaseModel):
    """Captures an exception raised when evaluating an :class:`ExampleOutput`.

    Attributes:
        error_message: String-representation of the exception.
    """

    error_message: str

    @staticmethod
    def from_exception(exception: Exception) -> "FailedExampleEvaluation":
        return FailedExampleEvaluation(
            error_message=f"{type(exception)}: {exception}\n{traceback.format_exc()}"
        )


class ExampleEvaluation(BaseModel, Generic[Evaluation]):
    """Evaluation of a single evaluated :class:`Example`.

    Created to persist the evaluation result in the repository.

    Attributes:
        evaluation_id: Identifier of the run the evaluated example belongs to.
        example_id: Identifier of the :class:`Example` evaluated.
        result: If the evaluation was successful, evaluation's result,
            otherwise the exception raised during running or evaluating
            the :class:`Task`.

    Generics:
        Evaluation: Interface of the metrics that come from the evaluated :class:`Task`.
    """

    evaluation_id: str
    example_id: str
    result: SerializeAsAny[Evaluation | FailedExampleEvaluation]

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return (
            f"Evaluation ID = {self.evaluation_id}\n"
            f"Example ID = {self.example_id}\n"
            f"Result = {self.result}\n"
        )

    def _rich_render(self, skip_example_id: bool = False) -> Tree:
        tree = Tree(f"Evaluation: {self.evaluation_id}")
        if not skip_example_id:
            tree.add(f"Example ID: {self.example_id}")
        tree.add(str(self.result))
        return tree


class PartialEvaluationOverview(BaseModel, frozen=True):
    """Overview of the un-aggregated results of evaluating a :class:`Task` on a dataset.

    Attributes:
        run_overviews: Overviews of the runs that were evaluated.
        id: The unique identifier of this evaluation.
        start: The time when the evaluation run was started.
        submitted_evaluation_count: The amount of evaluations that were submitted successfully.
        description: human-readable for the evaluator that created the evaluation.
    """

    run_overviews: frozenset[RunOverview]
    id: str
    start_date: datetime
    submitted_evaluation_count: int
    description: str
    labels: set[str] = set()
    metadata: SerializableDict = dict()

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        run_overview_str: str = "Run Overviews={\n"
        comma_counter = 0
        for overview in self.run_overviews:
            run_overview_str += f"{overview}"
            if comma_counter < len(self.run_overviews) - 1:
                run_overview_str += ", "
                comma_counter += 1
        run_overview_str += "}\n"

        return (
            f"Evaluation Overview ID = {self.id}\n"
            f"Start time = {self.start_date}\n"
            f"Submitted Evaluations = {self.submitted_evaluation_count}\n"
            f'Description = "{self.description}"\n'
            f"Labels = {self.labels}\n"
            f"Metadata = {self.metadata}\n"
            f"{run_overview_str}"
        )


class EvaluationOverview(BaseModel, frozen=True):
    """Overview of the un-aggregated results of evaluating a :class:`Task` on a dataset.

    Attributes:
        run_overviews: Overviews of the runs that were evaluated.
        id: The unique identifier of this evaluation.
        start_date: The time when the evaluation run was started.
        end_date: The time when the evaluation run was finished.
        successful_evaluation_count: Number of successfully evaluated examples.
        failed_evaluation_count: Number of examples that produced an error during evaluation.
            Note: failed runs are skipped in the evaluation and therefore not counted as failures
        description: human-readable for the evaluator that created the evaluation.
        labels: Labels for filtering evaluation. Defaults to empty list.
        metadata: Additional information about the evaluation. Defaults to empty dict.
    """

    run_overviews: frozenset[RunOverview]
    id: str
    start_date: datetime
    end_date: datetime
    successful_evaluation_count: int
    failed_evaluation_count: int
    description: str
    labels: set[str] = set()
    metadata: SerializableDict = dict()

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        run_overview_str: str = "Run Overviews={\n"
        comma_counter = 0
        for overview in self.run_overviews:
            run_overview_str += f"{overview}"
            if comma_counter < len(self.run_overviews) - 1:
                run_overview_str += ", "
                comma_counter += 1
        run_overview_str += "}\n"

        return (
            f"Evaluation Overview ID = {self.id}\n"
            f"Start time = {self.start_date}\n"
            f"End time = {self.end_date}\n"
            f"Successful examples = {self.successful_evaluation_count}\n"
            f"Failed examples = {self.failed_evaluation_count}\n"
            f'Description = "{self.description}"\n'
            f"Labels = {self.labels}\n"
            f"Metadata = {self.metadata}\n"
            f"{run_overview_str}"
        )

    def __hash__(self) -> int:
        return hash(self.id)


class EvaluationFailed(Exception):
    def __init__(self, evaluation_id: str, failed_count: int) -> None:
        super().__init__(
            f"Evaluation {evaluation_id} failed with {failed_count} failed examples."
        )

    end: datetime
    failed_example_count: int
    successful_example_count: int
    description: str



from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from typing import Optional

from pharia_studio_sdk.evaluation.aggregation.domain import (
    AggregatedEvaluation,
    AggregationOverview,
)


class AggregationRepository(ABC):
    """Base aggregation repository interface.

    Provides methods to store and load aggregated evaluation results: :class:`AggregationOverview`.
    """

    @abstractmethod
    def store_aggregation_overview(
        self, aggregation_overview: AggregationOverview[AggregatedEvaluation]
    ) -> None:
        """Stores an :class:`AggregationOverview`.

        Args:
            aggregation_overview: The aggregated results to be persisted.
        """
        ...

    @abstractmethod
    def aggregation_overview(
        self, aggregation_id: str, aggregation_type: type[AggregatedEvaluation]
    ) -> Optional[AggregationOverview[AggregatedEvaluation]]:
        """Returns an :class:`AggregationOverview` for the given ID.

        Args:
            aggregation_id: ID of the aggregation overview to retrieve.
            aggregation_type: Type of the aggregation.

        Returns:
            :class:`EvaluationOverview` if it was found, `None` otherwise.
        """
        ...

    def aggregation_overviews(
        self, aggregation_type: type[AggregatedEvaluation]
    ) -> Iterable[AggregationOverview[AggregatedEvaluation]]:
        """Returns all :class:`AggregationOverview`s sorted by their ID.

        Args:
            aggregation_type: Type of the aggregation.

        Yields:
            :class:`AggregationOverview`s.
        """
        for aggregation_id in self.aggregation_overview_ids():
            aggregation_overview = self.aggregation_overview(
                aggregation_id, aggregation_type
            )
            if aggregation_overview is not None:
                yield aggregation_overview

    @abstractmethod
    def aggregation_overview_ids(self) -> Sequence[str]:
        """Returns sorted IDs of all stored :class:`AggregationOverview`s.

        Returns:
            A :class:`Sequence` of the :class:`AggregationOverview` IDs.
        """
        pass



from abc import ABC, abstractmethod
from typing import Any, Optional

from pharia_inference_sdk.core import Input, Output
from pharia_inference_sdk.core.task import Task

from pharia_studio_sdk.evaluation.aggregation.aggregator import AggregationLogic
from pharia_studio_sdk.evaluation.aggregation.domain import AggregatedEvaluation
from pharia_studio_sdk.evaluation.dataset.domain import ExpectedOutput
from pharia_studio_sdk.evaluation.evaluation.domain import Evaluation
from pharia_studio_sdk.evaluation.evaluation.evaluator.evaluator import EvaluationLogic


class Benchmark(ABC):
    """Specific Benchmark instance used to run benchmark executions.

    Attributes:
        benchmark_id: Unique identifier for the benchmark.
        dataset_id: Identifier for the dataset used in the benchmark.
        eval_logic: Evaluation logic to be applied to task within the benchmark.
        aggregation_logic: Aggregation logic to combine individual evaluations into an aggregated result.
    """

    @abstractmethod
    def __init__(
        self,
        benchmark_id: str,
        dataset_id: str,
        eval_logic: EvaluationLogic[Input, Output, ExpectedOutput, Evaluation],
        aggregation_logic: AggregationLogic[Evaluation, AggregatedEvaluation],
        **kwargs: Any,
    ) -> None:
        """Initializes a Benchmark instance.

        Args:
            benchmark_id: Unique identifier for the benchmark.
            dataset_id: Identifier for the dataset associated with the benchmark.
            eval_logic: The evaluation logic to be applied.
            aggregation_logic: The aggregation logic to combine individual evaluations.
            **kwargs: Additional keyword arguments for extended configurations.
        """
        pass

    @abstractmethod
    def execute(
        self,
        task: Task[Input, Output],
        name: str,
        description: Optional[str],
        labels: Optional[set[str]],
        metadata: Optional[dict[str, Any]],
    ) -> str:
        """Executes the benchmark on a given task.

        Args:
            task: The task to be evaluated in the benchmark.
            name: Name of the benchmark execution.
            description: Description of the task to be evaluated.
            labels: Labels for filtering or categorizing the benchmark.
            metadata: Additional information about the task for logging or configuration.

        Returns:
            Identifier of the benchmark run.
        """
        pass


class BenchmarkRepository(ABC):
    """Used to manage Benchmark instances."""

    @abstractmethod
    def create_benchmark(
        self,
        dataset_id: str,
        eval_logic: EvaluationLogic[Input, Output, ExpectedOutput, Evaluation],
        aggregation_logic: AggregationLogic[Evaluation, AggregatedEvaluation],
        name: str,
        metadata: Optional[dict[str, Any]] = None,
        description: Optional[str] = None,
    ) -> Benchmark:
        """Creates a new benchmark and stores it in the repository.

        Args:
            dataset_id: Identifier for the dataset associated with the benchmark.
            eval_logic: Evaluation logic to be applied in the benchmark.
            aggregation_logic: Aggregation logic for combining individual evaluations.
            name: Name of the benchmark.
            metadata: Additional information about the benchmark, defaults to None.
            description: Description of the benchmark, defaults to None.

        Returns:
            The created benchmark instance.
        """
        pass

    @abstractmethod
    def get_benchmark(
        self,
        benchmark_id: str,
        eval_logic: EvaluationLogic[Input, Output, ExpectedOutput, Evaluation],
        aggregation_logic: AggregationLogic[Evaluation, AggregatedEvaluation],
        allow_diff: bool = False,
    ) -> Benchmark | None:
        """Retrieves an existing benchmark from the repository.

        Args:
            benchmark_id: Unique identifier for the benchmark to retrieve.
            eval_logic: Evaluation logic to apply.
            aggregation_logic: Aggregation logic to apply.
            allow_diff: Retrieve the benchmark even though logics behaviour do not match.

        Returns:
            The retrieved benchmark instance. Raises ValueError if no benchmark is found.
        """
        pass
